﻿using System.Diagnostics;
using System.Reflection.Metadata;
using System.Reflection.PortableExecutable;

namespace DotNetAssemblyFinder
{
    /// <summary>
    /// Entrypoint.
    /// </summary>
    internal class Program
    {
        /// <summary>
        /// Cache of parsed TargetFrameworkAttribute.
        /// </summary>
        private static readonly Dictionary<string, string> VER_DICT = [];

        /// <summary>
        /// Entrypoint.
        /// </summary>
        /// <param name="args">Command line arguments</param>
        /// <returns>Result</returns>
        internal static int Main(string[] args)
        {
            try
            {
                Options o = new(args);
                if (o.IsHelp)
                    DisplayUsage(0);
                if (o.Path == null)
                    DisplayUsage(-1);

                if (File.Exists(o.Path))
                {
                    Console.WriteLine(ReadMetadata(o.Path));
                    return 0;
                }
                else if (!Directory.Exists(o.Path))
                    DisplayUsage(-2);

                EnumerationOptions eo = new() { IgnoreInaccessible = true };
                if (o.IsRecursive)
                    eo.RecurseSubdirectories = true;

                int c = 0;
                Stopwatch sw = Stopwatch.StartNew();
                foreach (string s in Directory.EnumerateFiles(o.Path!, "*", eo))
                {
                    string? meta = ReadMetadata(s);
                    if (meta != null)
                        Console.WriteLine($"{++c}. {s}: {meta}");
                }
                
                if (o.IsDebug)
                    Console.Error.WriteLine($"Count: {c}, Elapsed: {sw.Elapsed}");
                    
                return 0;
            }
            catch (Exception e)
            {
                Console.Error.WriteLine(e.ToString());
                return int.MinValue;
            }
        }

        /// <summary>
        /// Read metadata from assembly file.
        /// </summary>
        /// <param name="file">Assembly file</param>
        /// <returns>Assembly target framework and version</returns>
        internal static string? ReadMetadata(string file)
        {
            try
            {
                using FileStream fs = new(file, FileMode.Open, FileAccess.Read);
                if (!fs.CanRead || fs.ReadByte() != 77 || fs.ReadByte() != 90) return null;
                fs.Position = 0;
                using PEReader pe = new(fs);
                if (!pe.HasMetadata) return null;
                var meta = pe.GetMetadataReader();
                if (meta == null) return null;
                return $"{GetTargetFramework(meta)}, v{meta.GetAssemblyDefinition().Version}";
            }
            catch (BadImageFormatException be)
            {
                Console.Error.WriteLine($"{file}: {be.Message}");
                return null;
            }
        }

        /// <summary>
        /// Get name and version of target framework.
        /// </summary>
        /// <param name="meta">Metadata</param>
        /// <returns>Framework name and version</returns>
        internal static string GetTargetFramework(MetadataReader meta)
        {
            string? ver = GetAttributeValue(meta, "TargetFrameworkAttribute");

            if (string.IsNullOrEmpty(ver)) return "Unknown";
            if (VER_DICT.TryGetValue(ver, out var value))
                return value;
            string v = ver[(ver.LastIndexOf('v') + 1)..ver.IndexOf('\u0001', 1)];
            value = ver[7] switch
            {
                'C' => $".NET Core {v}",
                'F' => $".NET Framework {v}",
                'S' => $".NET Standard {v}",
                _ => "Unknown",
            };
            VER_DICT.Add(ver, value);
            return value;
        }

        /// <summary>
        /// Get value of custom attribute.
        /// </summary>
        /// <param name="meta">Metadata</param>
        /// <param name="name">Attribute name</param>
        /// <returns>Attribute value</returns>
        internal static string? GetAttributeValue(MetadataReader meta, string name)
        {
            foreach (var h in meta.CustomAttributes)
            {
                try
                {
                    var a = meta.GetCustomAttribute(h);
                    var ctor = a.Constructor;
                    if (ctor.Kind != HandleKind.MemberReference) 
                        continue;

                    var container = meta.GetMemberReference((MemberReferenceHandle)ctor).Parent;
                    var tmp = meta.GetTypeReference((TypeReferenceHandle)container).Name;
                    if (meta.GetString(tmp) == name)
                    {
                        var reader = meta.GetBlobReader(a.Value);
                        return System.Text.Encoding.UTF8.GetString(reader.ReadBytes(reader.Length));
                    }
                }
                catch (InvalidCastException) { continue; }
            }
            return null;
        }

        /// <summary>
        /// Display error message and usage, and then exit.
        /// </summary>
        /// <param name="error">Error code (exit code)</param>
        internal static void DisplayUsage(int error)
        {
            switch (error)
            {
                case -1: Console.WriteLine("Error: No path."); break;
                case -2: Console.WriteLine("Error: Path not exist."); break;
            }
            Console.Write(@"Usage: DNAFinder [-d][-r] path

-d      Output debug messages.
-h      Display usage.
-r      Recursive to sub-folders.
path    File or Folder.
");
            Environment.Exit(error);
        }
    }
}
